import numpy as np
import re
import struct
import logging


def _get_binary(flag, binary: int):
    if binary == 4:
        if struct.unpack('>f', flag)[0] == 1234567.0:
            dtype = '>f4'
            dtype_str = '>f'
        elif struct.unpack('<f', flag)[0] == 1234567.0:
            dtype = '<f4'
            dtype_str = '<f'
        else:
            raise ValueError('Cannot determine binary 4 data dtype')

    elif binary == 8:
        if struct.unpack('>d', flag)[0] == 123456789012345.0:
            dtype = '>f8'
            dtype_str = '>d'
        elif struct.unpack('<d', flag)[0] == 123456789012345.0:
            dtype = '<f8'
            dtype_str = '<d'
        else:
            raise ValueError('Cannot determine binary 8 data dtype')

    else:
        raise ValueError('Unsupported binary data size')

    return dtype, dtype_str


class OVF:
    """Class for reading OVF files."""

    @classmethod
    def _validate(cls, filename: str):
        try:
            open(filename, 'rb')
        except FileNotFoundError:
            logging.warning("No such file or directory.")
            return False
        return True

    def __new__(cls, filename: str):
        if cls._validate(filename):
            return super().__new__(cls)

    def __init__(self, filename: str):
        #: Path to the ovf file.
        self.filename: str = filename
        #: Dictionary with OVF metadata.
        self.properties: dict = {}
        #: Magnetization read from the OVF file provided as a numpy array with shape (3, nx, ny, nz):
        #: (magnetization component, spatial x dimension, spatial y dimension, spatial z dimension).
        self.magnetization: np.ndarray
        self._get_data()

    def _get_data(self):
        """Loads metadata and magnetization binary from OVF file."""

        pattern = r'#\s(?!Begin|End)([\w]+): (\d.?\d*[eE][+-]\d+]?|\d+)'  # matches header metadata
        try:
            file = open(self.filename, 'rb')
        except FileNotFoundError:
            logging.warning("No such file or directory.")
        else:
            with file:
                for line in file:
                    try:
                        decoded_line = line.decode()[:-1]  # remove newline character
                    except UnicodeDecodeError:
                        logging.warning("Can't read line")
                    logging.debug(decoded_line)
                    content = re.findall(pattern, decoded_line)
                    if content:
                        name, value = content[0]
                        self.properties[name] = float(value)
                    elif decoded_line.startswith('# Begin: Data Binary '):
                        binary = int(decoded_line[-1])
                        break

                flag = file.read(binary)
                dtype, dtype_str = _get_binary(flag, binary)

                nx, ny, nz = map(int, [self.properties['xnodes'], self.properties['ynodes'], self.properties['znodes']])
                n_sites = nx * ny * nz
                data = np.fromfile(file, dtype=dtype)
                self.magnetization = np.array(
                    [data[i:3 * n_sites:3].reshape((nx, ny, nz), order='F') for i in range(3)])



import logging
import numpy as np
from typing import Union
from magpack.vectorop import cart2pol, sph2cart, cart2sph
import itertools
import random


def checkerboard(shape: tuple) -> np.ndarray:
    """Creates a checkerboard array of alternating +1 and -1 for the given shape.

    :param shape:   Shape of the checkerboard.
    :return:        Array of alternating +1 and -1 for the given shape.
    """
    return (np.indices(shape).sum(axis=0) % 2) * 2 - 1


def circ_mask(nx: int, ny: int) -> np.ndarray:
    """Defines a circular binary mask.

    :param nx:      x dimension of mask.
    :param ny:      y dimension of mask.
    :return:        Binary mask."""
    xx, yy = create_mesh(nx, ny)
    circ = np.sqrt(xx ** 2 + yy ** 2) < np.min([nx, ny]) // 2
    return circ


def vortex(nx: int, ny: int, winding: float = 1) -> np.ndarray:
    """Creates a 2D magnetization vortex.

    :param nx:      x spatial dimension of vortex.
    :param ny:      y spatial dimension of vortex.
    :param winding: Winding number of the vortex.
    :return:        Magnetization vector of vortex.
    """
    xx, yy = create_mesh(nx, ny)
    rad, azimuth = cart2pol(yy, xx)

    mx = -np.cos(winding * azimuth)
    my = np.sin(winding * azimuth)
    mz = np.zeros_like(azimuth)

    return np.array([mx, my, mz])


def domain_generator(shape: list, points: int, cart: bool = False) -> tuple[np.ndarray, list[tuple]]:
    """Generates Voronoi domains with random orientations.

    :param shape:   Shape of the final array.
    :param points:  Number of seed points.
    :param cart:    Resulting orientations will be provided in cartesian coordinates if True,
                    spherical coordinates if False.
    :return:        Array with orientations (3, shape) and list of seed point coordinates.
    """
    if np.prod(shape) <= points:
        raise ValueError('Number of seeds is greater than the number of points.')
    elements = list(map(np.arange, shape))
    grid = np.meshgrid(*elements, indexing='ij')
    all_randoms = itertools.product(*elements)
    randoms = sorted(all_randoms, key=lambda k: random.random())[:points]

    azimuths = np.random.rand(points) * 2 * np.pi
    elevations = np.random.rand(points) * np.pi
    min_distance_field = np.ones(tuple(shape)) * max(shape) * 2
    orientation_field = np.zeros((2,) + tuple(shape))
    for az, el, p_vect in zip(azimuths, elevations, randoms):
        difference_field = np.sqrt(sum([pow(dist - p, 2) for dist, p in zip(grid, p_vect)]))
        loc = np.where(difference_field < min_distance_field)
        orientation_field[0][loc] = az
        orientation_field[1][loc] = el
        min_distance_field[loc] = difference_field[loc]

    if cart:
        orientation_field = sph2cart(np.ones_like(orientation_field[0]), orientation_field[0], orientation_field[1])
    return orientation_field, randoms


def skyrmion(nx, ny, number=1, helicity=0, polarity=1, neel: bool = False) -> np.ndarray:
    """Creates a skyrmion texture of size (nx, ny).

    :param nx:          Number of points in the x-direction.
    :param ny:          Number of points in the y-direction.
    :param number:      The skyrmion topological number.
    :param helicity:    Global angular offset.
    :param polarity:    Direction of center spin (±1).
    :param neel:        Neel or Bloch skyrmion.
    :return:            Magnetization vector field of a skyrmion.
    """
    xx, yy = create_mesh(nx, ny)
    rad, azimuth = cart2pol(xx, -yy)

    # Normalize polarity to ±1 and neel to ±1
    polarity = np.sign(polarity)
    if neel and helicity:
        logging.warning("Neel skyrmions should not have a helicity.")

    theta = 2 * rad / (np.min([nx, ny]) - 1) * np.pi
    my, mx, mz = sph2cart(np.ones_like(theta), theta, number * (azimuth + helicity))

    if neel:
        my, mx = -mx, my

    mx = np.where(theta > np.pi, 0, mx)
    my = np.where(theta > np.pi, 0, my)
    mz = np.where(theta > np.pi, -polarity, mz) * polarity
    return np.array([mx, my, mz])


def meron(nx: int, ny: int, number: float = 1) -> np.ndarray:
    """Creates a magnetic meron.

    :param nx:      Number of points in the x-direction.
    :param ny:      Number of points in the y-direction.
    :param number:  The meron topological number.
    :return:        Magnetization vector field of a meron."""
    xx, yy = create_mesh(nx, ny)
    rad, azimuth = cart2pol(xx, -yy)
    theta = rad / (np.min([nx, ny]) - 1) * np.pi
    my, mx, mz = sph2cart(np.ones_like(theta), theta, number * azimuth)
    mz = np.where(theta > np.pi / 2, 0, mz)
    return np.array([mx, my, mz])


def domain_wall(nx: int, ny: int, neel=False, width=2):
    """Creates a 2D magnetic domain wall.

    :param nx:      Number of points in the x-direction.
    :param ny:      Number of points in the y-direction.
    :param neel:    Neel (if True) or Bloch (if False) type domain wall.
    :param width:   Width of the domain wall.
    :return:        Magnetization vector field of the domain wall.
    """

    xx, yy = create_mesh(nx, ny)
    angle = np.arctan(xx / width)
    mx = np.cos(angle)
    my = -np.sin(angle)
    mz = np.zeros_like(angle)

    if neel:
        mx, mz = mz, mx
    return np.array([mx, my, mz])


def bloch_point(nx, ny, nz, inwards=False, winding=1):
    """Creates a Bloch point topological defect.

    :param nx:          Number of points in the x-direction.
    :param ny:          Number of points in the y-direction.
    :param nz:          Number of points in the z-direction.
    :param inwards:     If True, magnetization points into the core.
    :param winding:     Winding number of the bloch defect.
    :return:            Magnetization vector field of a bloch point.
    """
    xx, yy, zz = create_mesh(nx, ny, nz)
    _, t, p = cart2sph(xx, yy, zz)
    mz = -np.cos(t) if inwards else np.cos(t)
    mx = np.sin(t) * np.sin(winding * p)
    my = -np.sin(t) * np.cos(winding * p)
    return np.array([mx, my, mz])


def meron_pair(nx: int, ny: int):
    """Creates a meron-antimeron pair.

    :param nx:      Number of points in the x-direction.
    :param ny:      Number of points in the y-direction.
    :return:        Magnetization vector field of a meron-antimeron pair.
    """
    mer = meron(nx, ny, 1)
    anti_mer = meron(nx, ny, -1)
    return np.hstack([mer, -anti_mer])


def stack_config(config, repeat, axis=-1):
    """Stacks 2D slices to form a 3D structure.

    :param config:  2D configuration of the structure.
    :param repeat:  Number of times to stack the configuration.
    :param axis:    Axis along which to stack the configuration.
    :return:        3D configuration."""
    return np.stack([config] * repeat, axis=axis)


def _create_mesh_ints(*args: list[int]):
    """Creates an ND mesh centered at the origin.

    :param args:  The integer dimensions of the mesh.
    :return:      ND mesh centered at the origin with unit spacing.
    """
    if any(not isinstance(v, int) for v in args) or min(args) < 1:
        raise ValueError("Only positive integer values can be converted to a mesh.")
    vectors = map(lambda x: np.arange(x) - (x - 1) / 2, args)
    return np.meshgrid(*vectors, indexing='ij')


def create_mesh(*args: Union[list[int], list[float]]):
    """Creates an ND mesh with the specified dimensions.

    :param args:  The integer dimensions of the mesh or all points along the dimension.
    :return:      ND mesh for the corresponding array."""
    if len(args) < 1:
        raise ValueError("Need more than one dimensions.")
    if all(isinstance(v, int) for v in args):
        logging.debug("all ints")
        return _create_mesh_ints(*args)
    return np.meshgrid(*args, indexing='ij')



import numpy as np
from magpack import _ovf_reader


def save_vtk(filename: str, scalar_dict: dict) -> None:
    """Saves a dictionary of scalar fields to a VTK file.

    :param filename:        Name of output file.
    :param scalar_dict:     Dictionary of scalar fields with the same shape.
    """
    from pyevtk.hl import gridToVTK
    shapes = [v.shape for v in scalar_dict.values() if isinstance(v, np.ndarray)]
    if any(shapes[0] != shape for shape in shapes):
        raise ValueError("All scalar fields must have the same shape.")

    x_size, y_size, z_size = shapes[0]
    x = np.arange(x_size)
    y = np.arange(y_size)
    z = np.arange(z_size)
    gridToVTK(filename, x, y, z, pointData=scalar_dict)


def save_mat(filename: str, **data_dictionary) -> None:
    """Saves function arguments to a .mat file. Wrapper for scipy.io.savemat.

    :param filename:            Name of output file.
    :param data_dictionary:     Unpacked dictionary of data."""
    from scipy.io import savemat
    savemat(filename, data_dictionary)


def load_mat(filename: str) -> dict:
    """Loads a .mat file. Wrapper for scipy.io.loadmat.

    :param filename:    Name of input file.
    :return:            Dictionary of loaded variables."""
    from scipy.io import loadmat
    return loadmat(filename)


def load_ovf(filename: str) -> _ovf_reader.OVF:
    """Loads a .ovf file and return an OVF object.

    The magnetization can be accessed using OVF.magnetization and metadata using the OVF.properties.

    :param filename:    Name of input file.
    :return:            OVF object."""
    return _ovf_reader.OVF(filename)


def see_keys(data: dict, prefix: str = '') -> None:
    """Recursively prints keys of a dictionary. Useful for HDF5 files.

    :param data:    Data (e.g. an HDF5 file).
    :param prefix:  Prefix to prepend to keys."""
    try:
        keys = list(data.keys())
    except AttributeError:
        return None

    for j in keys:
        previous = prefix + j
        print(previous)
        see_keys(data[j], previous + '/')


def pil_save(img: np.array, filename: str, cmap: str = 'viridis', vmin: float = None, vmax: float = None,
             alpha: bool = False, alpha_thresh: int = 750, indexing: str = 'ij') -> None:
    """Saves a numpy array as a full resolution png file.

    :param img:             Array to be saved.
    :param filename:        Name of output file.
    :param cmap:            Matplotlib colormap name.
    :param vmin:            Lower bound for colorbar axis (defaults to minimum value in the img array).
    :param vmax:            Upper bound for colorbar axis (defaults to maximum value in the img array).
    :param alpha:           Option to make bright pixels (white) transparent.
    :param alpha_thresh:    Threshold value for transparency (max 765 = 255*3).
    :param indexing:        Indexing scheme (xy is for matplotlib convention, default is ij).
    """
    import matplotlib as mpl
    from PIL import Image

    vmin = img.min() if vmin is None else vmin
    vmax = img.max() if vmax is None else vmax

    img = np.flip(img.T, axis=0) if indexing == 'ij' else img

    img = np.clip(img, vmin, vmax)
    img = (img - vmin) / (vmax - vmin)
    c = mpl.colormaps[cmap]
    save_im = c(img) * 255
    if alpha:
        mask = np.sum(save_im, -1) >= alpha_thresh
        save_im[mask, -1] = 0
    save_im = np.uint8(save_im)
    save_im = Image.fromarray(save_im)
    save_im.save(filename)



import numpy as np
import logging
import re
from functools import reduce
from typing import Union


def _validate_rotation(matrix_shape: tuple, vf_shape: tuple) -> None:
    """Checks if the matrix shape is congruent with the shape of the vector field.

    :param matrix_shape:    Shape of the matrix.
    :param vf_shape:        Shape of the vector field.
    """
    if not len(vf_shape) > 1:
        raise ValueError(f"Vector field must be at least two dimensional. Given field has shape {vf_shape}.")
    if len(matrix_shape) != 2:
        raise ValueError(f"Rotation matrix must be two-dimensional. Given matrix has shape {matrix_shape}.")
    elif matrix_shape[1] != vf_shape[0]:
        raise ValueError(f"Matrix dimensions do not match the number of components. Matrix has shape {matrix_shape},"
                         f" field has shape {vf_shape}.")
    if not matrix_shape[0] == matrix_shape[1]:
        logging.warning(f"Non-square rotation matrix, information will be lost.")


def rotx(theta: float, degrees: bool = True) -> np.ndarray:
    """Generates the 3x3 rotation matrix for a rotation about the x-axis.

    :param theta:       Angle of rotation.
    :param degrees:     True for degrees, False for radians.
    :return:            Rotation matrix with shape(3, 3).
    """
    theta = np.deg2rad(theta) if degrees else theta

    c, s = np.cos(theta), np.sin(theta)
    r = np.array([[1, 0, 0],
                  [0, c, -s],
                  [0, s, c]])
    return r


def roty(theta: float, degrees: bool = True) -> np.ndarray:
    """Generates the 3x3 rotation matrix for a rotation about the y-axis.

    :param theta:       Angle of rotation.
    :param degrees:     True for degrees, False for radians.
    :return:            Rotation matrix with shape(3, 3).
    """
    theta = np.deg2rad(theta) if degrees else theta

    c, s = np.cos(theta), np.sin(theta)
    r = np.array([[c, 0, s],
                  [0, 1, 0],
                  [-s, 0, c]])
    return r


def rotz(theta: float, degrees: bool = True) -> np.ndarray:
    """Generates the 3x3 rotation matrix for a rotation about the z-axis.

    :param theta:       Angle of rotation.
    :param degrees:     True for degrees, False for radians.
    :return:            Rotation matrix with shape(3, 3).
    """
    theta = np.deg2rad(theta) if degrees else theta

    c, s = np.cos(theta), np.sin(theta)
    r = np.array([[c, -s, 0],
                  [s, c, 0],
                  [0, 0, 1]])
    return r


def rot(theta: float, degrees: bool = True):
    """Generates the 2x2 rotation matrix for an in-plane rotation.

    :param theta:       Angle of rotation.
    :param degrees:     True for degrees, False for radians.
    :return:            Rotation matrix with shape(2, 2).
    """
    theta = np.deg2rad(theta) if degrees else theta
    c, s = np.cos(theta), np.sin(theta)

    r = np.array([[+c, -s],
                  [+s, +c]])
    return r


def eul2rot(seq: str, *args, degrees: bool = True) -> np.ndarray:
    r"""Returns a stack of rotation matrices following the sequence and the corresponding angles.

    :param seq:         Sequence of rotation operations (first operation listed first).
    :param args:        Unpacked list of angles (or array of angles) matching the length of the sequence.
    :param degrees:     True for degrees, False for radians.
    :return:            Stack of rotation matrices with shape(n, 3, 3). `n` is the product of the length of all args.

    Examples
    ----------
    Dual-axis tomography: Describes full 180° rotations for each of the two tilts: 0° and 30°.

    >>> eul2rot('zy', [0, 30], np.linspace(1, 180))

    Laminography: Full 360° rotation around an axis that is slanted by 45° around the y-axis.

    >>> eul2rot('yz', np.linspace(1,360) ,45)

    """
    seq_regex = re.compile("^[xyz]+$")
    if not (seq_regex.match(seq)):
        raise ValueError("Sequence must be a combination of 'xyz' only.")

    if not degrees:
        args = [np.rad2deg(arg) for arg in args]

    len_seq = len(seq)
    n_args = len(args)

    if len(args) != len(seq):
        raise ValueError(f"Sequence of rotations must match input arguments. Sequence has {len_seq} operations but"
                         f" {n_args} were given.")
    all_rotations = np.meshgrid(*args, indexing='ij')
    table = np.array(all_rotations).T.reshape(-1, n_args, order='F')

    # creates a 2D table with all operations. Each row is one measurement orientation.
    rot_dict = {'x': rotx, 'y': roty, 'z': rotz}
    rot_all = map(lambda z: reduce(lambda x, y: x @ y, [rot_dict[axis](angle) for axis, angle in zip(seq, z)]), table)
    return np.array(list(rot_all)).squeeze()


def tomo_rot(angles: np.ndarray, tilts: Union[float, np.ndarray] = 0) -> np.ndarray:
    r"""Generates a stack of rotation matrices describing tomography with tilting.

    :param angles:      Angles of rotation within each tilt series.
    :param tilts:       Array of tilts.
    :return:            Stack of rotation matrices shaped (n, 3, 3). `n` is the product of the size of `angles` and
                        `tilts`.
    """
    return eul2rot('zy', tilts, angles)


def lamni_rot(angles: np.ndarray, lamni_tilt: Union[float, np.ndarray] = 45, lamni_skew: np.ndarray = 0) -> np.ndarray:
    r"""Generates a stack of rotation matrices describing laminography.

    :param angles:      Angles of rotation within each tilt series.
    :param lamni_tilt:  Array of laminography tilts.
    :param lamni_skew:  Array of skewness in laminography.
    :return:            Stack of rotation matrices shaped (n, 3, 3). `n` is the product of the size of `angles`,
                        `lamni_tilt` and `lamni_skew`.
    """
    return eul2rot('yzx', angles, lamni_tilt, lamni_skew)


def rotate_field(vector_field: np.ndarray, rot_matrix: np.ndarray) -> np.ndarray:
    """Rotates each element in the vector field according to the provided rotation matrix.

    The vector field must have the form (n, x, y, ..., z) where n indexes the component.

    :param vector_field:        N-dim vector field, shape (n, x, y, ..., z).
    :param rot_matrix:          Rotation matrix, shape (m, n).
    :return:                    Rotated N-dim vector field, shape (m, x, y, ..., z).
    """

    _validate_rotation(rot_matrix.shape, vector_field.shape)
    return np.einsum('ij,j...->i...', rot_matrix, vector_field)


def rotate_vector_field(vector_field: np.ndarray, rot_matrix: np.ndarray, order: int = 1) -> np.ndarray:
    """Rotates a vector field according to the provided rotation matrix.

    :param vector_field:        Vector field to be rotated, shape (3, x, y, z).
    :param rot_matrix:          Rotation matrix, shape (3, 3).
    :param order:               Interpolation order (0 to 5, default is 1).
    :return:                    Rotated vector field, shape (3, x, y, z).
    """
    # add another dimension to the rotation matrix and send to scalar recons
    n_spatial_dims = vector_field.ndim - 1
    vector_field = rotate_field(vector_field, rot_matrix)

    new_rot_matrix = np.zeros((n_spatial_dims + 1, n_spatial_dims + 1))
    new_rot_matrix[0, 0] = 1
    new_rot_matrix[1:, 1:] = rot_matrix[0:n_spatial_dims, 0:n_spatial_dims]

    return rotate_scalar_field(vector_field, new_rot_matrix, order=order)


def rotate_scalar_field(field: np.ndarray, rot_matrix: np.ndarray, order: int = 1) -> np.ndarray:
    """Rotates a scalar field according to the provided rotation matrix.

    :param field:        Scalar field to be rotated, shape (x, y, z).
    :param rot_matrix:   Rotation matrix, shape (3, 3).
    :param order:        Interpolation order (0 to 5, default is 1).
    :return:             Rotated scalar field, shape (x, y, z).
    """
    from scipy.ndimage import affine_transform
    if not (0 <= order <= 5):
        raise ValueError(f"Interpolation order must be an integer between 0 and 5. Given order is {order}.")
    vf_shape = np.asarray(field.shape)

    rot_matrix = rot_matrix.T  # the rotation matrix is transposed to match the affine_transform convention

    out_center = rot_matrix @ (vf_shape - 1) / 2
    in_center = (vf_shape - 1) / 2
    offset = in_center - out_center
    output = affine_transform(field, rot_matrix, offset=offset, order=order)
    return output



from typing import Union, Any

import numpy as np
from itertools import combinations

from numpy import ndarray, dtype, signedinteger

DEFAULT_INTERPOLATION_ORDER = 1


def scalar_gradient(scalar_field: np.ndarray) -> np.ndarray:
    """Compute the gradient of a scalar field.

    :param scalar_field: Scalar field to compute the gradient.
    :return:             (Vector) gradient of scalar field.
    """
    return np.stack(np.gradient(scalar_field))


def vector_laplacian(vector_field: np.ndarray) -> np.ndarray:
    """Compute the laplacian of a vector field (cartesian).

    :param vector_field: Input vector field.
    :return:             Vector laplacian of input field.
    """
    return np.stack([scalar_laplacian(i) for i in vector_field])


def scalar_laplacian(scalar_field: np.ndarray) -> np.ndarray:
    """Calculates the scalar Laplacian of a scalar field.

    :param scalar_field: Scalar field to calculate the Laplacian.
    :return: (Vector)    Laplacian of scalar field."""
    # get all spatial dimension gradients, with the first index being the component
    return divergence(scalar_gradient(scalar_field))


def divergence(vector_field: np.ndarray) -> np.ndarray:
    """Calculates the divergence of a vector field.

    :param vector_field: Input vector field.
    :return: (Vector)    Divergence of input field."""
    return np.add.reduce([np.gradient(vector_field[i], axis=i) for i in range(vector_field.shape[0])])


def levi_civita_nd(*args) -> np.ndarray:
    """Calculate N-Dimensional Levi Civita tensor

    :param args:    Dimensions of tensor.
    :return:        N-Dimensional Levi Civita tensor."""

    def __levi_civita(*sub_args) -> Union[ndarray[Any, dtype[Any]], signedinteger[Any]]:
        if len(sub_args) != len(set(sub_args)):
            return np.array(0)
        combs = combinations(reversed(sub_args), 2)
        signs = [np.sign(x - y) for x, y in combs]
        return np.prod(signs)

    vec_lc = np.vectorize(__levi_civita)
    return vec_lc(*args)


def _levi_civita(i: np.ndarray, j: np.ndarray, k: np.ndarray) -> np.ndarray:
    """Calculates the value of (i,j,k) element of the Levi-Civita tensor.

    :param i: Index i of the array.
    :param j: Index j of the array.
    :param k: Index k of the array.
    :return:  Corresponding Levi-Civita tensor."""
    return (i - j) * (j - k) * (k - i) / 2


def vorticity(vector_field: np.ndarray) -> np.ndarray:
    r"""Calculates the magnetic vorticity vector field for the given vector field.

    :param vector_field: Three-dimensional vector field shaped (3, nx, ny, nz) for which the vorticity will be
                         calculated.
    :return:             Three-dimensional vorticity vector field shaped (3, nx, ny, nz).

    The magnetic vorticity, :math:`\mathbf{\Omega}`, of the magnetization vector field, :math:`\mathbf{m}`, is given by
    [1]_, [2]_:

    .. math::

        \Omega_i = \frac{1}{8\pi}\epsilon_{abc}\epsilon_{ijk}m_{i}\partial_{b}m_{j}\partial_{c}m_{k}

    References
    ----------
    .. [1] Papanicolaou, N. , Tomaras, T. N. Nucl. Phys. B 360, 2-3
    .. [2] Cooper, N. R. Phys. Rev. Lett. 82, 1554

    """
    epsilon = _levi_civita(*np.indices((3, 3, 3)))
    diffs = np.stack([np.gradient(vector_field[i]) for i in range(vector_field.shape[0])])
    vort = np.einsum('ijk,abc,jbxyz,kcxyz,ixyz->axyz', epsilon, epsilon, diffs, diffs, vector_field)
    return vort / (8 * np.pi)


def skyrmion_number(vector_field: np.ndarray) -> np.ndarray:
    r"""Calculates the skyrmion topological number.

    The skyrmion topological number for the vector field :math:`\mathbf{m}` is given by evaluating integral:

    .. math::

        \frac{1}{4\pi} \int \mathbf{m} \cdot \left( \frac{\partial \mathbf{m}}{\partial x} \times \frac{\partial
        \mathbf{m}}{\partial y} \right)\,\text{d}x \text{d}y


    :param vector_field:    The vector field shape (n, x, y) from which the skyrmion number will be calculated.
    :return:                The skyrmion number.
    """
    components = vector_field.shape[0]
    spatial_dimensions = vector_field.ndim - 1

    if components != 3:
        raise ValueError("Vector field must have 3 components.")
    if spatial_dimensions != 2:
        raise ValueError("Field must have 2 spatial dimensions. ")

    vector_field = normalize(vector_field)
    epsilon = _levi_civita(*np.indices((3, 3, 3)))

    # differentials for x and y components
    diffs = np.stack([np.gradient(vector_field[i]) for i in range(3)])
    sk_n = np.einsum('ixy,ijk,jxy,kxy->xy', vector_field, epsilon, diffs[:, 0], diffs[:, 1])

    return sk_n.sum() / (4 * np.pi)


def curl(vector_field: np.ndarray) -> np.ndarray:
    """Calculates the curl of a vector field.

    :param vector_field: Vector field shape (3, nx, ny, nz) from which the curl will be calculated.
    :return:             Curl of vector field."""

    epsilon = _levi_civita(*np.indices((3, 3, 3)))
    diffs = np.stack([np.gradient(vector_field[i]) for i in range(vector_field.shape[0])])
    return np.einsum('ijk,jkabc->iabc', epsilon, diffs)


def magnitude(vector_field: np.ndarray) -> np.ndarray:
    """Calculates the magnitude of a vector field.

    :param vector_field:    Vector field shape (3, nx, ny, nz).
    :return:                Magnitude of vector field."""

    return np.sqrt(np.sum(vector_field ** 2, axis=0))


def scale_range(values: np.ndarray, norm_list: list = None, mask_zero: bool = True) -> np.ndarray:
    """Scales all values in the array such that they lie within the specified range.

    :param values:      Array to be scaled.
    :param norm_list:   List of length 2 with lower and upper bound (defaults to [-1, 1]).
    :param mask_zero:   If True, zero values will remain zero..
    :return:            Scaled values."""
    if norm_list is None:
        norm_list = [-1, 1]
    norm_range = norm_list[1] - norm_list[0]

    vmin, vmax = np.min(values), np.max(values)
    vrange = vmax - vmin
    values = (values - vmin) / vrange * norm_range + norm_list[0]

    if mask_zero:
        mask = np.where(values == 0, 0, 1)
        values = values * mask
    return values


def normalize(vector_field: np.ndarray) -> np.ndarray:
    """Scales all vectors in the array to unit length.

    :param vector_field:    The vector field to normalize.
    :return:                The normalized vector field."""
    mag = magnitude(vector_field)
    return np.divide(vector_field, mag, where=mag != 0)


def cart2sph(x: np.ndarray, y: np.ndarray, z: np.ndarray) -> np.ndarray:
    """Converts cartesian coordinates to spherical coordinates.

    All arrays must have the same shape (a, b, c). The resulting array has shape (3, a, b, c) and can be unpacked to the
    radial, elevation and azimuthal components.

    :param x:   The x cartesian coordinates.
    :param y:   The y cartesian coordinates.
    :param z:   The z cartesian coordinates.
    :return:    Corresponding spherical coordinates."""

    r = np.sqrt(x ** 2 + y ** 2 + z ** 2)
    th = np.arccos(z / r, where=r != 0)
    ph = np.arctan2(y, x, where=np.logical_and(x != 0, y != 0))
    return np.stack([r, th, ph])


def sph2cart(r: np.ndarray, th: np.ndarray, ph: np.ndarray) -> np.ndarray:
    """Converts spherical coordinates to cartesian coordinates.

    All arrays must have the same shape (x, y, z). The resulting array has shape (3, a, b, c) and can be unpacked to the
    x, y and z components.

    :param r:       The spherical radial coordinates.
    :param th:      The spherical elevation coordinates.
    :param ph:      The spherical azimuthal coordinates.
    :return:        Corresponding cartesian coordinates."""

    x = r * np.sin(th) * np.cos(ph)
    y = r * np.sin(th) * np.sin(ph)
    z = r * np.cos(th)
    return np.stack([x, y, z])


def cart2pol(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """Converts 2D cartesian coordinates to polar coordinates.

    All arrays must have the same shape (a, b). The resulting array has shape (2, a, b) and can be unpacked to the
    radial and azimuthal components.

    :param x:   The x cartesian coordinates.
    :param y:   The y cartesian coordinates.
    :return:    Corresponding polar coordinates."""

    r = np.sqrt(x ** 2 + y ** 2)
    az = np.arctan2(y, x)
    return np.stack([r, az])


def angular_gradient(data: np.ndarray, axial=False) -> np.ndarray:
    """Calculates the angular gradient of a vector field.
    
    :param data:    The vector field to calculate the angular gradient from.
    :param axial:   True for orientation fields, False for directional fields.
    :return:        The angular gradient scalar field."""
    delta_theta = [angular_difference(data, np.roll(data, 1, axis=i), axial=axial) for i in range(1, data.ndim)]
    delta_theta = np.sum(np.array(delta_theta), axis=0) / 3
    return delta_theta


def angular_difference(in_a: np.ndarray, in_b: np.ndarray, axial=False) -> np.ndarray:
    """Calculates the angular difference between vectors at the same positions of array a and b.

    :param in_a:    The first vector field to compare.
    :param in_b:    The second vector field to compare.
    :param axial:   True for orientation fields, False for directional fields.
    :return:        The scalar field of angular difference between the two inputs."""

    if in_a.shape != in_b.shape:
        raise ValueError("Inputs must have the same shape.")
    mag_mul = magnitude(in_a) * magnitude(in_b)
    dot_product = np.sum(in_a * in_b, axis=0)
    dot_product = np.where(dot_product > mag_mul, mag_mul, dot_product)  # normalise
    if axial:
        dot_product = np.abs(dot_product)
    np.seterr(invalid='ignore')
    ang_diff = np.arccos(dot_product / mag_mul, where=mag_mul != 0, out=np.zeros_like(dot_product))
    # replace nans with 0
    ang_diff = np.where(np.isnan(ang_diff), 0, ang_diff)
    return ang_diff


def magnitude_difference(in_a: np.ndarray, in_b: np.ndarray, percent=True) -> np.ndarray:
    """Calculates the magnitude difference between vectors at the same positions of array a and b.

    :param in_a:    The first vector field to compare.
    :param in_b:    The second vector field to compare.
    :param percent: True to calculate percentage difference, False to calculate absolute difference.
    :return:        The scalar field of magnitude difference between the two inputs.
    """

    m_a = magnitude(in_a)
    m_b = magnitude(in_b)
    if in_a.shape != in_b.shape:
        raise ValueError("Inputs must have the same shape.")
    if percent:  # if percent is used then in_b is considered to be the reference
        return 100 * np.divide(m_a - m_b, m_b, where=m_b != 0, out=np.zeros_like(m_a))
    else:
        return m_a - m_b


def stokes_to_jones(stokes: np.ndarray) -> np.ndarray:
    """Converts Stokes polarization vector to Jones polarization vector.

    :param stokes:  The Stokes polarization vector, shape (4,).
    :return:        The Jones polarization vector.
    """
    dop = np.sqrt(np.sum(np.power(stokes[1:], 2)))
    horizontal = stokes[1] / dop
    diagonal = stokes[2] / dop
    circular = stokes[3] / dop

    a = np.sqrt((1 + horizontal) / 2)
    if a == 0:
        b = 1
    else:
        b = diagonal / (2 * a) - 1j * circular / (2 * a)
    return np.sqrt(dop) * np.array([a, b])


def jones_to_stokes(jones: np.ndarray) -> np.ndarray:
    """ Converts Jones polarization vector to Stokes polarization vector.

    :param jones:  The Jones polarization vector, shape (2,).
    :return:        The Stokes polarization vector."""
    jones = jones / np.sqrt(np.sum(np.abs(jones) ** 2))
    m = jones[0] * jones[1].conjugate()
    return np.array([1, np.abs(jones[0]) ** 2 - np.abs(jones[1]) ** 2, 2 * np.real(m), 2 * np.imag(m)])
